'' Edit Map 1 - Map Editor
''
'' Author: MSrobots
'' Version: 2
'' Date: 2018-09-08
''
'' Copyright (c) 2018 Michael Sommer
'' See end of file for terms of use.
''
CON                                 
                                                                              
  _clkmode = XTAL1|PLL16X                                                                    
  _xinfreq = 5_000_000

  gravity = 100 'The rate in which the spaceship descents in free fall (in ten seconds) 
  maxstep = 9 ' numbers of gridlines
  
OBJ
  wfs    : "WFsystem"           ' Basic WF support

VAR
  long state[4]                 'State of each Game
  long oldstate[4]              'old State for HELP and TOOLS

  long curMapF[4]               'address of filename of map
  long curMapW[4]               'width   in map for HELP and TOOLS
  long curMapH[4]               'height  in map for HELP and TOOLS
  long curMapX[4]               'xpos    in map for HELP and TOOLS
  long curMapY[4]               'ypos    in map for HELP and TOOLS
  long curMapC[4]               'Color   in map for HELP and TOOLS
  
  long wrkMapF[4]               'address of filename of WORK map ? needed? should it be the same?
  long wrkMapW[4]               'width   in WORK map
  long wrkMapH[4]               'height  in WORK map
  long wrkMapX[4]               'xpos    in WORK map
  long wrkMapY[4]               'ypos    in WORK map
  long wrkMapC[4]               'Color   in WORK map
  long wrkCurX[4]               'CursorX in WORK map
  long wrkCurY[4]               'CursorY in WORK map

  byte wrkMode[4]               'DrawMode: 0 Move Cursor, 1 Draw Stamp on leaving
  byte wrkCurCh[4]              'Cursor Char Index into -\|/ at @CursorChars
  byte wrkCurFB[4]              'Cursor ForeBack Color
  byte wrkStampCh[4]            'Stamp Char
  byte wrkStampFB[4]            'Stamp ForeBack Color

  byte selX[4]                  'current select cursor X
  byte selY[4]                  'current select cursor Y
    
DAT '                                ###   THE MAIN LOOP   ###
PUB main | s, ch, sw, sh, i
's=screen, i/j=gp, s0/s1=secs time 0/1, ch=char, rot=rotate, 'ch=kbd character
  wfs.initSystem     'start up system: video, keyboards, sd card, etc.
  splashScreens
  repeat s from 0 to 3
    startEditor(s)
  
  repeat 'main loop, infinite
    s := (s + 1) & 3       'cycle among the keyboards (0..3, wrapping around)

    if state[s]==1
      drawCursor(s, false)
    elseif state[s]==3
      incCursor(s)
      wfs.pokeColor(s, selX[s],selY[s], wrkCurFB[s])
    elseif state[s]==4
      incCursor(s)
      wfs.pokeChar(s, 12+selX[s],1+selY[s], byte[@CursorChars+wrkCurCh[s]])
      wfs.pokeColor(s, 12+selX[s],1+selY[s], wrkCurFB[s])
      
    ch := wfs.getKey(s)            'get key in keyboard buffer, if present; 0 = no key
    
    if ch <> 0                       'check if there's a buffered keystroke for curr. kbd.
        
      if (state[s] < 1) and (ch <> 203)
        if state[s]<0             ' State <0 restarts and sets state to 0
          startEditor(s)
        else
          state[s] := 1           ' State0 wird State1
          drawScreen(s)           ' output new screen 
          drawState(s, true)      ' output current values with cursor positions
      elseif state[s]==1          
        State1(s, ch)             ' showing EDITOR
      elseif state[s]==2          
        State2(s, ch)             ' showing HELP or FILE
      elseif state[s]==3          
        State3(s, ch)             ' showing CHARPICK
      else
        State4(s, ch)             ' showing COLPICK
                                                                                                  
PUB version
RETURN 1

DAT                           'State Support Methods for MAIN LOOP
PRI State1(s, ch) | i            ' showing EDITOR
  case ch                 
    203:
      wfs.bootPartition(string("/GAMEMENU/GAMEMENU.BIN")) 'launch GameMenu
    208:
      'F1 - Help
      oldstate[s] :=  state[s]
      state[s] := 2
      StoreTempS := s+48
      wfs.storeScreen(s, @StoreTemp,-2)
      curMapF[s] := @HelpFile
      curMapC[s] := wfs#RG' file contains just Text, so Colors are fixed
      curMapW[s] := 42
      curMapH[s] := wfs.fileSize(@HelpFile) / curMapW[s]
      curMapX[s] := 0
      curMapY[s] := 0
      drawCur(s)
    209:
      'F2 - Move/Draw/Dchr/Dcol
       wrkMode[s]++
       if wrkMode[s]>6
         wrkMode[s] := 0
      if wrkMode[s] > 0
        repeat i from 0 to 39
          wfs.pokeColor(s, i,0, wfs#KR)
          wfs.pokeColor(s, i,1, wfs#KR)
          wfs.pokeColor(s, i,14, wfs#KR)
      else
        repeat i from 0 to 39
          wfs.pokeColor(s, i,0, wfs#KB)
          wfs.pokeColor(s, i,1, wfs#KB)
          wfs.pokeColor(s, i,14, wfs#KB)
      drawModeValues(s)  
      drawStampValues(s)
      wfs.beep  
    210:
      'F3 - Char
      oldstate[s] :=  state[s]
      state[s] := 3
      StoreTempS := s+48
      wfs.storeScreen(s, @StoreTemp,-2)
      wfs.restoreScreen(s, @CharPick,-2)
      selY[s] := wrkStampCh[s] >> 4
      selX[s] := wrkStampCh[s] & 15
      if selY[s] >7
        selY[s] -= 8
        selX[s] += 22
      else
        selX[s] += 3
      selY[s] += 4
      wfs.pokeColor(s, selX[s],selY[s], wfs#RK)
      
    211:
      'F4 - Col
      oldstate[s] :=  state[s]
      state[s] := 4
      StoreTempS := s+48
      wfs.storeScreen(s, @StoreTemp,-2)
      
      wfs.blitMap(s, @ColorPick, 14, 14, 0, 0, 14, 14, 12, 1, -2, 12, 0) 

      selY[s] := 0#> (wrkStampFB[s] >> 4) <#7
      selX[s] := 0#> (wrkStampFB[s] & 15) <#7
      selX[s] += 3
      selY[s] += 3
      wfs.pokeChar(s, 12+selX[s],1+selY[s], "X")
    212:
      'F5 - Pick
      removeCursor(s, ch) 
      wrkStampCh[s] := wfs.peekChar(s, wrkCurX[s] - wrkMapX[s], wrkCurY[s] - wrkMapY[s] + 2)
      wrkStampFB[s] := wfs.peekColor(s, wrkCurX[s] - wrkMapX[s], wrkCurY[s] - wrkMapY[s] + 2)
      drawStampValues(s)
      drawCursor(s, false)
      wfs.beep  
    213:
      'F6 - File
      oldstate[s] :=  state[s]
      state[s] := 2
      StoreTempS := s+48
      wfs.storeScreen(s, @StoreTemp,-2)
      curMapF[s] := @ToolsMap
      curMapC[s] := -2 ' file contains Colors
      curMapW[s] := 42
      curMapH[s] := (wfs.fileSize(@ToolsMap) / curMapW[s]) >> 1
      curMapX[s] := 0
      curMapY[s] := 0
      drawCur(s)
      ' Work Map
    232:  'up
      removeCursor(s, ch) 
      if (wrkCurY[s]>wrkMapY[s]) and (wrkCurY[s]>0)
        wrkCurY[s]--
        drawCursor(s, true)
      else
        if wrkMapY[s] > -11
          wrkMapY[s]--
          if wrkCurY[s]>0
            wrkCurY[s]--
            drawState(s, true)
          else
            drawState(s, false)
    228: 'left
      removeCursor(s, ch) 
      if (wrkCurX[s]>wrkMapX[s]) and (wrkCurX[s]>0)
        wrkCurX[s]--
        drawCursor(s, true)
      else
        if wrkMapX[s] > -39
          wrkMapX[s]--
          if wrkCurX[s]>0
            wrkCurX[s]--
            drawState(s, true)
          else
            drawState(s, false)
    230:  'right
      removeCursor(s, ch) 
      if (wrkCurX[s]<(wrkMapX[s]+40-1)) and (wrkCurX[s]<(wrkMapW[s]-1))
        wrkCurX[s]++
        drawCursor(s, true)
      else
        if wrkMapX[s] < (wrkMapW[s]-1)
          wrkMapX[s]++
          if wrkCurX[s]<(wrkMapX[s]-1)
            wrkCurX[s]++
            drawState(s, true)
          else
            drawState(s, false)
    226:  'down          
      removeCursor(s, ch) 
      if (wrkCurY[s]<(wrkMapY[s]+12-1)) and (wrkCurY[s]<(wrkMapH[s]-1))
        wrkCurY[s]++
        drawCursor(s, true)
      else
        if wrkMapY[s] < (wrkMapH[s]-1)
          wrkMapY[s]++ 
          if wrkCurY[s]<(wrkMapH[s]-1)
            wrkCurY[s]++
            drawState(s, true)
          else
            drawState(s, false)
    other:
      if (wrkMode[s]>0) and (wrkMode[s]<4)
        removeCursor(s, ch) 
        if (wrkCurX[s]<(wrkMapX[s]+40-1)) and (wrkCurX[s]<(wrkMapW[s]-1))
          wrkCurX[s]++
          drawCursor(s, true)
        else
          if wrkMapX[s] < (wrkMapW[s]-1)
            wrkMapX[s]++
            if wrkCurX[s]<(wrkMapX[s]-1)
              wrkCurX[s]++
              drawState(s, true)
            else
              drawState(s, false)
    

PRI State2(s, ch) | sw, sh      ' showing HELP or FILE
  sw := 34
  sh := 11
  case ch
    232:  'up
      if curMapY[s] > -(sh-1)
        curMapY[s]--
        drawCur(s)
    228: 'left
      if curMapX[s] > -(sw-1)
        curMapX[s]--
        drawCur(s)
    230:  'right
      if curMapX[s] < (curMapW[s]-1)
        curMapX[s]++
        drawCur(s)
    226:  'down          
      if curMapY[s] < (curMapH[s]-1)
        curMapY[s]++ 
        drawCur(s)
    other:      
      state[s] := oldstate[s]
      StoreTempS := s+48
      wfs.restoreScreen(s, @StoreTemp,-2)
      drawState(s, false)
   
PRI State3(s, ch)  | curChar, curColor, resColor ' showing CHARPICK
  curChar := byte[@CursorChars+wrkCurCh[s]]
  curColor := wrkCurFB[s]
  resColor := 16*(selY[s] - 3)
  resColor += (selX[s] - 3)  
  case ch
    232:  'up
      wfs.pokeColor(s, selX[s],selY[s], wfs#BW)
      if selY[s] > 4
        selY[s]--
      else
        selY[s] := 11
      wfs.pokeColor(s, selX[s],selY[s], curColor)
    228:  'left
      wfs.pokeColor(s, selX[s],selY[s], wfs#BW)
      if selX[s]<21
        if selX[s] > 3
          selX[s]--
        else
          selX[s] := 37
      else
        if selX[s] > 22
          selX[s]--
        else
          selX[s] := 18
      wfs.pokeColor(s, selX[s],selY[s], curColor)
    230:  'right
      wfs.pokeColor(s, selX[s],selY[s], wfs#BW)
      if selX[s]<21
        if selX[s] > 17
          selX[s] := 22
        else
          selX[s]++
      else
        if selX[s] > 36
          selX[s] := 3
        else
          selX[s]++
      wfs.pokeColor(s, selX[s],selY[s], curColor)
    226:  'down          
      wfs.pokeColor(s, selX[s],selY[s], wfs#BW)
      if selY[s] > 10
        selY[s] := 4
      else
        selY[s]++
      wfs.pokeColor(s, selX[s],selY[s], curColor)
    13:   'enter
      if selX[s]<21
        wrkStampCh[s] := 16*(selY[s] - 4)
        wrkStampCh[s] += (selX[s] - 3)
      else
        wrkStampCh[s] := 16*(selY[s] + 4)
        wrkStampCh[s] += (selX[s] - 22)
        
      state[s] := oldstate[s]
      StoreTempS := s+48
      wfs.restoreScreen(s, @StoreTemp,-2)
      drawState(s, false)
      drawStampValues(s)
      wfs.beep  
    other:      
      state[s] := oldstate[s]
      StoreTempS := s+48
      wfs.restoreScreen(s, @StoreTemp,-2)
      drawState(s, false)

PRI State4(s, ch) | curChar, curColor, resColor ' showing COLPICK
  curChar := byte[@CursorChars+wrkCurCh[s]]
  curColor := wrkCurFB[s]
  resColor := 16*(selY[s] - 3)
  resColor += (selX[s] - 3)  
  case ch
    232:  'up
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, "X")
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, resColor)
      if selY[s] > 3
        selY[s]--
      else
        selY[s] := 10
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, curChar)
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, curColor)
    228:  'left
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, "X")
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, resColor)
      if selX[s] > 3
        selX[s]--
      else
        selX[s] := 10
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, curChar)
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, curColor)
    230:  'right
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, "X")
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, resColor)
      if selX[s] > 9
        selX[s] := 3
      else
        selX[s]++
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, curChar)
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, curColor)
    226:  'down          
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, "X")
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, resColor)
      if selY[s] > 9
        selY[s] := 3
      else
        selY[s]++
      wfs.pokeChar(s, selX[s]+12,selY[s]+1, curChar)
      wfs.pokeColor(s, selX[s]+12,selY[s]+1, curColor)
    13:   'enter
      wrkStampFB[s] := resColor  
      state[s] := oldstate[s]
      StoreTempS := s+48
      wfs.restoreScreen(s, @StoreTemp,-2)
      drawState(s, false)
      drawStampValues(s)
      wfs.beep  
    other:      
      state[s] := oldstate[s]
      StoreTempS := s+48
      wfs.restoreScreen(s, @StoreTemp,-2)
      drawState(s, false)

DAT                           'Support Methods for MAIN LOOP
PRI splashScreens | s, i 
  'show splash momentarily
  wfs.beep
  StoreScreensN := "0"
  repeat s from 0 to 3
    wfs.restoreScreen(s, @StoreScreens,-2)
  waitcnt(clkfreq*5+cnt)
  wfs.beep
                                    
PRI startEditor(s) | row
  wrkMapF[s] := @EditFile'@EditFile ' @CharPick '@ColorPick 
  wrkMapC[s] := -2 ' file contains Colors
  wrkMapW[s] := 407'  40' 407 '40 '14
  wrkMapH[s] := (wfs.fileSize(wrkMapF[s]) / wrkMapW[s]) >> 1
  wrkMapX[s] := 0
  wrkMapY[s] := 0
  wrkCurX[s] := 0
  wrkCurY[s] := 0
  wrkMode[s] := 0
  wrkCurCh[s] := 0
  wrkCurFB[s] := 0
  wrkStampCh[s] := 32
  wrkStampFB[s] := 32
  
  state[s] := 0   

  StoreScreensN := "1"
  wfs.restoreScreen(s, @StoreScreens,-2)

PRI drawScreen(s)
  StoreScreensN := "2"
  wfs.restoreScreen(s, @StoreScreens,-2)
  drawModeValues(s) 
  drawStampValues(s)
  
PRI drawState(s, posCursorChanged) 
  wfs.blitMap(s, wrkMapF[s], wrkMapW[s], wrkMapH[s], wrkMapX[s], wrkMapY[s], 40, 12, 0, 2, wrkMapC[s], ".", wfs#WK)
  drawCursor(s, posCursorChanged)

PRI drawCur(s) 
  wfs.blitMap(s, curMapF[s], curMapW[s], curMapH[s], curMapX[s], curMapY[s], 40, 13, 0, 2, curMapC[s], ".", wfs#WK)
  
PRI drawNumber6(s, c, r, k, n, nosign) | col, val, ch, sign
  val := n
  if val<0
    sign := "-"
    val := -val
  else
    sign := nosign
  col := c + 5
  repeat 5
    ch := (val // 10) + 48
    val := val / 10
    wfs.putChar(s, col, r, k, ch)
    col--
  wfs.putChar(s, col, r, k, sign)

PRI drawHex4(s, c, r, k, n) | col, val, ch
  val := n
  col := c+4
  repeat 4
    ch := (val & 15)
    if ch > 9
      ch += "A"-10
    else
      ch += "0"
    val := val >> 4
    wfs.putChar(s, col, r, k, ch)
    col--

PRI drawModeValues(s)
  if wrkMode[s] == 0
    wfs.write(s, 12, 0, wfs#KB, @Move)
  elseif wrkMode[s] == 1
    wfs.write(s, 12, 0, wfs#BR, @Type)
  elseif wrkMode[s] == 2
    wfs.write(s, 12, 0, wfs#BR, @Tchr)
  elseif wrkMode[s] == 3
    wfs.write(s, 12, 0, wfs#BR, @Tcol)
  elseif wrkMode[s] == 4
    wfs.write(s, 12, 0, wfs#GR, @Draw)
  elseif wrkMode[s] == 5
    wfs.write(s, 12, 0, wfs#GR, @Dchr)
  elseif wrkMode[s] == 6
    wfs.write(s, 12, 0, wfs#GR, @Dcol)

PRI drawStampValues(s) | FB 
  if wrkMode[s] == 0
    FB := wfs#KB
  else
    FB := wfs#KR
  drawHex4(s, 32, 1, FB,(wrkStampFB[s] << 8 ) + wrkStampCh[s])
  wfs.putChar(s, 38, 1, wrkStampFB[s], wrkStampCh[s])

PRI incCursor(s) | curColorF, curColorB
  curColorF := wrkCurFB[s] >> 4 
  curColorB := (wrkCurFB[s] & 15) +1
  if curColorB >7
    curColorB := 0
    curColorF += 1
    if curColorF >7
      curColorF := 0
      wrkCurCh[s] += 1
      if wrkCurCh[s]>3
        wrkCurCh[s] := 0     
  wrkCurFB[s] := (curColorF<<4)+curColorB
  
PRI drawCursor(s, posCursorChanged) | FB, scrFB, scrCH 
  if wrkMode[s] == 0
    FB := wfs#KB
  else
    FB := wfs#KR
  incCursor(s)  
  if posCursorChanged
    drawNumber6(s, 16, 14, FB, wrkCurX[s], ":")
    drawNumber6(s, 24, 14, FB, wrkCurY[s], ":")
    scrCH := wfs.peekChar(s, wrkCurX[s] - wrkMapX[s],wrkCurY[s] - wrkMapY[s] + 2)
    scrFB := wfs.peekColor(s, wrkCurX[s] - wrkMapX[s],wrkCurY[s] - wrkMapY[s] + 2)
    drawHex4(s, 32, 14, FB,(scrFB<<8) + scrCH)
    wfs.putChar(s, 38, 14, scrFB, scrCH)
  wfs.putChar(s, wrkCurX[s] - wrkMapX[s], wrkCurY[s] - wrkMapY[s] + 2, wrkCurFB[s], byte[@CursorChars+wrkCurCh[s]])

PRI removeCursor(s, ch) | i, offset, Mode
    ' wrkMode[s] == 0 just MOVE
    ' neuen wert in file eintragen,
    ' wrkMode[s] == 1 TYPE - both txt and color
    ' wrkMode[s] == 2 TCHR - just txt
    ' wrkMode[s] == 3 TCOL - just color
    ' wrkMode[s] == 4 DRAW - both txt and color
    ' wrkMode[s] == 5 DCHR - just txt
    ' wrkMode[s] == 6 DCOL - just color    
  Mode := wrkMode[s]
  if Mode > 0 and Mode < 4
    case ch
      203:  'ESC
        Mode := 0
      208..213:
        'F1 - Help
        'F6 - Pick
        Mode := 0
      232:  'up
        Mode := 0
      228: 'left
        Mode := 0
      230:  'right
        Mode := 0
      226:  'down          
        Mode := 0
      other:
        wrkStampCh[s] := ch
        drawStampValues(s)
        Mode += 3      
  wfs.blitMap(s, wrkMapF[s], wrkMapW[s], wrkMapH[s], wrkCurX[s], wrkCurY[s], 1, 1,wrkCurX[s] - wrkMapX[s], wrkCurY[s] - wrkMapY[s] + 2, wrkMapC[s], ".", wfs#WK)
  if Mode > 0
    if Mode == 4
      offset := wrkMapW[s]*wrkMapH[s]                   'offset positiv both text and color
    elseif Mode == 5
      offset :=0                                        'offset 0 just text
    elseif Mode == 6
      offset := -wrkMapW[s]*wrkMapH[s]                  'offset negativ, just color   
    wfs.pokeFile(wrkMapF[s], (wrkMapW[s]*wrkCurY[s])+wrkCurX[s], wrkStampCh[s], wrkStampFB[s], offset)
    repeat i from 0 to 3                                ' dann fuer jeden screen aktualisieren
      if state[i]==1                                    ' Map visible
        if wrkCurX[s]>wrkMapX[i]-1                      ' not left of map[i]
          if wrkCurY[s]>wrkMapY[i]-1                    ' not above of map[i]
            if wrkMapX[i]+40>wrkCurX[s]                 ' not right of map[i] 
              if wrkMapY[i]+12>wrkCurY[s]               ' not below of map[i]
                if (Mode==4) or (Mode==5)
                  wfs.pokeChar(i, wrkCurX[s] - wrkMapX[i], wrkCurY[s] - wrkMapY[i] + 2, wrkStampCh[s])
                if (Mode==4) or (Mode==6)
                  wfs.pokeColor(i, wrkCurX[s] - wrkMapX[i], wrkCurY[s] - wrkMapY[i] + 2, wrkStampFB[s])
                wfs.beep                                ' beep that you now you are changing things ?
     
DAT
Move          byte "MOVE ",0
Draw          byte "DRAW ",0
Dchr          byte "DCHR ",0
Dcol          byte "DCOL ",0
Type          byte "TYPE ",0
Tchr          byte "TCHR ",0
Tcol          byte "TCOL ",0
ExitStart     byte " ESC to EXIT - any other key to START   ",0
CursorChars   byte "-\|/"
StoreScreens  byte "/EDITMAP1/STORESCR.0"
StoreScreensN byte "0","4",0
StoreTemp     byte "/EDITMAP1/STORETMP.00"
StoreTempS    byte "0",0
HelpFile      byte "/EDITMAP1/EDITMAP1.HLP",0
ToolsMap      byte "/EDITMAP1/TOOLS000.MAP",0
CharPick      byte "/EDITMAP1/CHARPICK.MAP",0
ColorPick     byte "/EDITMAP1/COLPICK0.MAP",0
EditFile      byte "/EDITMAP1/GAME0000.MAP",0
'EditFile      byte "/LANDER02/SCOREMAP.MAP",0
'EditFile      byte "/LANDER02/STORESCR.014",0

DAT
{{

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                  TERMS OF USE: MIT License
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
// files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
// modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
// Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
// WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}}                